import logging

from sqlalchemy import Connection, ClauseElement

from alchemica.log_explain_plan import log_explain_plan
from alchemica.utils.utils import compile_sql_query


def before_execute(
        conn: Connection,
        clauseelement: ClauseElement,
        logger: logging.Logger,
        compile_sql: bool,
        explain_plan: bool,
        execution_info: str,
):
    compiled_query = None
    if compile_sql:
        try:
            compiled_query = compile_sql_query(conn=conn, clauseelement=clauseelement)
            logger.info(f"🟢 [SQL]{execution_info} {compiled_query}")
        except Exception as err:
            logger.warning(f"⚠️ [SQL]{execution_info} Ошибка при компиляции запроса {err}")
    else:
        logger.info(f"🟢 [SQL]{execution_info} {clauseelement}")

    if explain_plan:
        try:
            log_explain_plan(
                conn=conn,
                compiled_query=compiled_query,
                clauseelement=clauseelement,
                logger=logger,
                execution_info=execution_info
            )
        except Exception as err:
            logger.warning(f"⚠️ [SQL PLAN]{execution_info} Ошибка при получении плана запроса {err}")



from sqlalchemy.engine.base import Connection


def _invoke_before_exec_event(
        self,
        elem,
        distilled_params,
        execution_options,
):
    from sqlalchemy import exc

    if len(distilled_params) == 1:
        event_multiparams, event_params = [], distilled_params[0]
    else:
        event_multiparams, event_params = distilled_params, {}

    for fn in list(self.dispatch.before_execute):
        elem, event_multiparams, event_params = fn(
            self,
            elem,
            event_multiparams,
            event_params,
            execution_options,
        )

    if event_multiparams:
        distilled_params = list(event_multiparams)
        if event_params:
            raise exc.InvalidRequestError(
                "Event handler can't return non-empty multiparams "
                "and params at the same time"
            )
    elif event_params:
        distilled_params = [event_params]
    else:
        distilled_params = []

    return elem, distilled_params, event_multiparams, event_params


def patch_connection_invoke_before_exec_event():
    Connection._invoke_before_exec_event = _invoke_before_exec_event



import logging


def get_logger() -> logging.Logger:
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)  # Устанавливаем уровень логирования
    console_handler = logging.StreamHandler()  # Создаем обработчик для вывода в консоль
    console_handler.setLevel(logging.INFO)  # Устанавливаем уровень для консольного вывода
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(formatter)  # Применяем форматирование
    logger.addHandler(console_handler)  # Добавляем обработчик к логгеру
    return logger



import logging

from sqlalchemy import Connection, text, ClauseElement

from alchemica.utils.utils import compile_sql_query


def get_explain_plan(
        conn: Connection,
        compiled_query,
):
    explain_query = text(f"EXPLAIN {compiled_query}")
    result = conn.execute(explain_query)
    plan = result.fetchall()
    plan = "\n".join(str(i[0]) for i in plan)
    return plan


def log_explain_plan(
        conn: Connection,
        compiled_query,
        clauseelement: ClauseElement,
        logger: logging.Logger,
        execution_info: str,
):
    if compiled_query is None:
        compiled_query = compile_sql_query(conn=conn, clauseelement=clauseelement)
    plan = get_explain_plan(conn=conn, compiled_query=compiled_query)
    logger.info(f"📝 [SQL PLAN]{execution_info} \n{plan}")



import logging
from contextvars import ContextVar
from functools import wraps
from time import time
from typing import Optional

from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError

from alchemica.after_execute import after_execute
from alchemica.before_execute import before_execute
from alchemica.utils.default_logger import get_logger
from alchemica.utils.sqlalchemy_patch import patch_connection_invoke_before_exec_event
from alchemica.utils.utils import get_uuid

__all__ = ["sync_sql_logger", ]

execution_context_var = ContextVar("execution_context_var")
sql_request_start_time_var = ContextVar("execution_start_time_var")
execution_info_var = ContextVar("execution_info_var")

patch_connection_invoke_before_exec_event()


def sync_sql_logger(
        logger: Optional[logging.Logger] = get_logger(),
        log_level: int = logging.INFO,
        log_execution_info: bool = True,
        compile_sql: bool = False,
        critical_time: Optional[float] = 1,
        explain_plan: bool = False,
        uuid_len: int = 6,
):
    logger.setLevel(log_level)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            def _before_execute(conn, clauseelement, multiparams, params, execution_options):  # noqa
                if execution_context_var.get(None) != func_uuid:
                    return
                if str(clauseelement).startswith("EXPLAIN"):
                    return
                query_uuid = get_uuid(uuid_len=uuid_len)
                if log_execution_info:
                    execution_info = f"(func_name: {func_name}; func_uuid: {func_uuid}; query_uuid: {query_uuid})"
                else:
                    execution_info = ""
                execution_info_var.set(execution_info)
                before_execute(
                    conn=conn,
                    clauseelement=clauseelement,
                    logger=logger,
                    compile_sql=compile_sql,
                    explain_plan=explain_plan,
                    execution_info=execution_info
                )
                sql_request_start_time_var.set(time())

            def _after_execute(conn2, clauseelement, multiparams, params, execution_options, result):  # noqa
                if execution_context_var.get(None) != func_uuid:
                    return
                if str(clauseelement).startswith("EXPLAIN"):
                    return
                execution_info = execution_info_var.get(None)
                sql_request_start_time = sql_request_start_time_var.get()
                after_execute(
                    critical_time=critical_time,
                    logger=logger,
                    execution_info=execution_info,
                    sql_request_start_time=sql_request_start_time,
                )

            func_uuid = get_uuid(uuid_len=uuid_len)
            previous_execution_context_var = execution_context_var.get(None)
            execution_context_var.set(func_uuid)

            sync_engine = Engine
            event.listen(sync_engine, "before_execute", _before_execute)
            event.listen(sync_engine, "after_execute", _after_execute)

            func_name = func.__name__
            if log_execution_info:
                func_execution_info = f"(func_name: {func_name}; func_uuid: {func_uuid})"
            else:
                func_execution_info = ""

            func_start_time = time()
            logger.info(f"🚀 [FUNC]{func_execution_info} запущена")
            try:
                func_result = func(*args, **kwargs)
                current_time = time()
                elapsed_time = current_time - func_start_time
                logger.info(f"✅ [FUNC]{func_execution_info} выполнена за {elapsed_time:.4f} сек")
            except SQLAlchemyError as err:
                logger.exception(f"❌ [SQL ERROR]{func_execution_info} {str(err)}")
                raise err
            except Exception as err:
                logger.exception(f"❌ [FUNC ERROR]{func_execution_info} {func_name} {str(err)}")
                raise err
            finally:
                execution_context_var.set(previous_execution_context_var)

                event.remove(sync_engine, "before_execute", _before_execute)
                event.remove(sync_engine, "after_execute", _after_execute)
            return func_result

        return wrapper

    return decorator



import logging
import time


def after_execute(
        critical_time: float,
        logger: logging.Logger,
        execution_info: str,
        sql_request_start_time: float,
):
    if sql_request_start_time is None:
        return
    elapsed_time = time.time() - sql_request_start_time
    logger.info(f"🕑️ [SQL]{execution_info} Выполнено за {elapsed_time:.4f} сек")
    if critical_time and elapsed_time > critical_time:
        logger.warning(
            f"⚠️ [SQL]{execution_info} Запрос выполнялся дольше {critical_time} секунд: {elapsed_time:.4f} сек"
        )



import uuid

from sqlalchemy import Connection, ClauseElement


def compile_sql_query(
        conn: Connection,
        clauseelement: ClauseElement,
):
    compiled_query = clauseelement.compile(
        dialect=conn.dialect,
        compile_kwargs={"literal_binds": True}  # Подставляем реальные значения
    )
    return compiled_query


def get_uuid(uuid_len: int = 6) -> str:
    result = uuid.uuid4().hex[-uuid_len:]
    return result


async def atime():
    import asyncio
    result = asyncio.get_event_loop().time()
    return result
